/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2016                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/
/*! \file trimesh_base.cpp
\ingroup code_sample

\brief the minimal example of using the lib

This file contain a minimal example of the library, showing how to load a mesh and how to compute per vertex normals on it.

*/

#include<vcg/complex/complex.h>
#include<vcg/complex/algorithms/update/color.h>
#include<wrap/io_trimesh/import_off.h>
#include<wrap/io_trimesh/import_obj.h>

#include<wrap/io_trimesh/export_off.h>
#include<wrap/io_trimesh/export_ply.h>
#include<vcg/complex/algorithms/clean.h>

using namespace vcg;

class MyVertex; class MyEdge; class MyFace;
struct MyUsedTypes : public vcg::UsedTypes<vcg::Use<MyVertex>   ::AsVertexType,
                                           vcg::Use<MyEdge>     ::AsEdgeType,
                                           vcg::Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public vcg::Vertex< MyUsedTypes, vcg::vertex::Coord3f, vcg::vertex::Normal3f, vcg::vertex::Color4b, vcg::vertex::Qualityf, vcg::vertex::BitFlags  >{};
class MyFace    : public vcg::Face<   MyUsedTypes, vcg::face::FFAdj,  vcg::face::VertexRef, vcg::face::Color4b, vcg::face::BitFlags > {};
class MyEdge    : public vcg::Edge<   MyUsedTypes> {};

class MyMesh    : public vcg::tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace> , std::vector<MyEdge>  > {};

// Given a face and an edge index, add a quad to the pathMesh
// the quad is placed in the middle of the edge

void AddQuadPerEdge(MyMesh &pathMesh, MyMesh::FacePointer fp, int i, float w0=0.5f)
{
  float w1=(1.0f-w0)/2.0f;
  Point3f V0 = fp->V0(i)->P()*w0 + fp->V1(i)->P()*w1 + fp->V2(i)->P()*w1;
  Point3f V1 = fp->V0(i)->P()*w1 + fp->V1(i)->P()*w0 + fp->V2(i)->P()*w1;
  MyMesh::FacePointer fp2=fp->FFp(i);
  int j=fp->FFi(i);
  Point3f V2 = fp2->V0(j)->P()*w0 + fp2->V1(j)->P()*w1 + fp2->V2(j)->P()*w1;
  Point3f V3 = fp2->V0(j)->P()*w1 + fp2->V1(j)->P()*w0 + fp2->V2(j)->P()*w1;
  
  tri::Allocator<MyMesh>::AddFace(pathMesh,V0,V2,V1);
  tri::Allocator<MyMesh>::AddFace(pathMesh,V0,V3,V2);
}

int main( int argc, char **argv )
{
  if(argc<2)
  {
    printf("Usage trimesh_base <meshfilename.obj>\n");
    return -1;
  }
  /*!
    */
  MyMesh m,pathMesh,wallMesh;
  srand(time(NULL));

  if(vcg::tri::io::ImporterOFF<MyMesh>::Open(m,argv[1])!=vcg::tri::io::ImporterOFF<MyMesh>::NoError)
  {
    printf("Error reading file  %s\n",argv[1]);
    exit(0);
  }

  tri::RequirePerVertexNormal(m);
  tri::UpdateNormal<MyMesh>::PerVertexNormalized(m);
  printf("Input mesh  vn:%i fn:%i\n",m.VN(),m.FN());
  vcg::tri::io::ExporterPLY<MyMesh>::Save(m,"input.ply", vcg::tri::io::Mask::IOM_VERTCOLOR|tri::io::Mask::IOM_VERTQUALITY); 

  tri::UpdateTopology<MyMesh>::FaceFace(m);
  
  std::vector<MyMesh::FacePointer> stack;
  tri::UpdateFlags<MyMesh>::FaceClearV(m);
  // sourceH is a Handle that allow to retrieve a typed attribute associated to each face. 
  // To use it just use the handle as a vector indexed with the object, 
  // e.g. to retrieve the attribute associated to a face f you can do
  // sourceH[f] (e.g. with the object)
  // sourceH[&f] (eg. with the pointer to the object)
  // sourceH[i] (e.g. with the index of the face in the vector of the faces m.face )
  auto sourceH = tri::Allocator<MyMesh>::AddPerFaceAttribute<MyMesh::FacePointer>(m,"source");
  stack.push_back(&m.face[0]);
  m.face[0].SetV();
  while(!stack.empty())
  {
    if(stack.size()>2)
      std::swap(stack[rand()%stack.size()],stack.back());

    MyMesh::FacePointer fp=stack.back();
    stack.pop_back();
    assert(fp->IsV());
    for(int i=0;i<3;++i)
    {
      if(!fp->FFp(i)->IsV())
      {
        fp->FFp(i)->SetV();
        stack.push_back(fp->FFp(i));
        AddQuadPerEdge(pathMesh,fp,i,0.7f);
        sourceH[fp->FFp(i)]=fp;
      }
      else
      {
        if(sourceH[fp]!=fp->FFp(i))
          AddQuadPerEdge(wallMesh,fp,i,0.7f);
      }
    }    
  }
  vcg::tri::io::ExporterPLY<MyMesh>::Save(pathMesh,"output.ply", vcg::tri::io::Mask::IOM_VERTCOLOR|tri::io::Mask::IOM_VERTQUALITY); 

  float w0=0.7f;
  float w1=(1.0f-w0)/2.0f;

  // Now add to pathMesh the missing small triangles
  for(size_t i=0;i<m.face.size();++i)
  {
    MyMesh::FacePointer fp=&m.face[i];
    Point3f V1 = fp->V(0)->P()*w1 + fp->V(1)->P()*w0 + fp->V(2)->P()*w1;
    Point3f V0 = fp->V(0)->P()*w0 + fp->V(1)->P()*w1 + fp->V(2)->P()*w1;
    Point3f V2 = fp->V(0)->P()*w1 + fp->V(1)->P()*w1 + fp->V(2)->P()*w0;
    tri::Allocator<MyMesh>::AddFace(pathMesh,V0,V1,V2);
  }
    
  // TODO Add the missing romboidal piece for each vertex of each triangle

  // TODO 2 (Optional) Move the wall faces up (use the face Normals fp->N() 
  // or the vertex Normals vp->N() but remember to initalize them)


// remove the duplicate vertice that we have created in the pathmesh
tri::Clean<MyMesh>::RemoveDuplicateVertex(pathMesh);
tri::Clean<MyMesh>::RemoveDuplicateVertex(wallMesh);

vcg::tri::io::ExporterPLY<MyMesh>::Save(pathMesh,"pathMesh.ply", vcg::tri::io::Mask::IOM_FACECOLOR); 
vcg::tri::io::ExporterPLY<MyMesh>::Save(wallMesh,"wallMesh.ply", vcg::tri::io::Mask::IOM_FACECOLOR); 

  return 0;
}
